
/* ----------------------------------------------------------------------
   The SINUS Firewall -- a TCP/IP packet filter for Linux

   Written within the SINUS project at the University of Zurich,
   SWITCH, Telekurs Payserv AG, ETH Zurich.
   originally based on the sf Firewall Software (C) 1996 by Robert
   Muchsel and Roland Schmid.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   SINUS Firewall resources:
   SINUS Homepage: http://www.ifi.unizh.ch/ikm/SINUS/
   Firewall Homepage: http://www.ifi.unizh.ch/ikm/SINUS/firewall.html
   Frequently asked questions: http://www.ifi.unizh.ch/ikm/SINUS/sf_faq.html
   Mailing list for comments, questions, bug reports: firewall@ifi.unizh.ch
   ----------------------------------------------------------------------  */



package sfclasses;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;

/**
 * This class implements notification levels. An object of this
 * class holds all information about level. The static parts of
 * the class take care of the level vectors and of static levels.
 * Static levels are stored in the vector <i>staticLevels</i> and are
 * inserted into that vector in a static initializer. Dynamic levels
 * are stored in a vector in the ManageDomain class.
 * @version 1.0 03 Dec 1996
 * @author Roland E. Schmid
 */

public class Notification implements Persistent {

  /**
   * Initialize empty level. This is used when loading data from a
   * persistent input stream.
   */
  protected Notification() {
  }

  /**
   * Initialize level.
   * @param number Level number
   */
  public Notification(int number) {
    levelNumber = number;
  }

  /**
   * Open a dialog to edit the notification level.
   * @param co Calling object, will be notified when the edit dialog is closed.
   * @param p Parent frame
   * @param md ManageDomain object holding the configuration data
   * @param newLevel The level is being created and not changed. If the user
   * cancels the operation, the level will be deleted from the vector.
   * @see Refreshable
   */
  public void edit(Refreshable co, Frame p, ManageDomain md, boolean newLevel) {
    NotificationDialog nd = new NotificationDialog(co, p, this, md, newLevel);
    nd.setVisible(true);
    if (levelNumber < firstDynamic)
      UserDialog.WarningBox("Static notification level. Changes will not be saved!");
  }

  /**
   * Print the notification level in configuration file format
   * @param ps PrintWriter to write the output to
   */
  public void printLevel(PrintWriter ps) {
    ps.println("level "+levelNumber+": /* "+levelName+" */");
    ps.println(levelText);
    ps.println();
  }

  // persistence

  /**
   * Write object data to a persistent output stream
   * @param ps Stream
   * @see PersistentOutputStream
   */
  public void write(PersistentOutputStream ps) {
    ps.writeInt("Level=", levelNumber);
    ps.writeString("levelName=", levelName);
    ps.writeString("levelText=", levelText);
  }

  /**
   * Read object data from a persistent input stream
   * @param ps Stream
   * @see PersistentInputStream
   */
  public void read(PersistentInputStream ps) throws java.io.IOException {
    levelNumber = ps.readInt("Level=");
    levelName = ps.readString("levelName=");
    levelText = ps.readString("levelText=");
  }

  protected String levelText = new String();
  protected int levelNumber;
  protected String levelName = new String();

  // static structures
  // =================

  /**
   * @param number number of notification level
   * @return notification level
   */
  public static Notification staticLevel(int number) {
    Enumeration en = staticLevels.elements();
    Notification n;
    while (en.hasMoreElements()) {
      n = (Notification)en.nextElement();
      if (n.levelNumber == number)
        return n;
    }
    return null;
  }

  /**
   * @param name name of notification level
   * @return notification level
   */
  public static Notification staticLevel(String name) {
    Enumeration en = staticLevels.elements();
    Notification n;
    while (en.hasMoreElements()) {
      n = (Notification)en.nextElement();
      if (n.levelName.equals(name))
        return n;
    }
    return null;
  }

  public static Notification getLevel(String name, Vector dynLevels) {
    Notification n = staticLevel(name);
    if (n == null) {
      Enumeration en = dynLevels.elements();
      while (en.hasMoreElements()) {
        n = (Notification)en.nextElement();
        if (n.levelName.equals(name))
          return n;
      }
    }
    else
      return n;
    return null;
  }

  public static Notification getLevel(int number, Vector dynLevels) {
    Notification n = staticLevel(number);
    if (n == null) {
      Enumeration en = dynLevels.elements();
      while (en.hasMoreElements()) {
        n = (Notification)en.nextElement();
        if (n.levelNumber == number)
          return n;
      }
    }
    else
      return n;
    return null;
  }

  public static void constructList(List l, Vector dynLevels, int levelnum) {
    Enumeration en = staticLevels.elements();
    Notification n;
    while (en.hasMoreElements()) {
      n = (Notification)en.nextElement();
      l.add(n.levelName);
      if (n.levelNumber == levelnum) {
        l.select(l.getItemCount()-1);
        l.makeVisible(l.getItemCount()-1);
      }
    }
    en = dynLevels.elements();
    while (en.hasMoreElements()) {
      n = (Notification)en.nextElement();
      l.add(n.levelName);
      if (n.levelNumber == levelnum) {
        l.select(l.getItemCount()-1);
        l.makeVisible(l.getItemCount()-1);
      }
    }
  }

  public static void constructList(List l, Vector dynLevels) {
    constructList(l, dynLevels, 0);
  }

  public static void printAllLevels(PrintWriter ps, Vector dynLevels) {
    Enumeration en = staticLevels.elements();
    Notification n;
    while (en.hasMoreElements()) {
      n = (Notification)en.nextElement();
      n.printLevel(ps);
    }
    en = dynLevels.elements();
    while (en.hasMoreElements()) {
      n = (Notification)en.nextElement();
      n.printLevel(ps);
    }
  }

  public static final int STATIC_SOURCEROUTE = 1;

  public static final int STATIC_SERVER      = 10;
  public static final int STATIC_IDENT       = 11;
  public static final int STATIC_TRACEROUTE  = 12;
  public static final int STATIC_OUTGOING    = 13;
  public static final int STATIC_PING        = 14;
  public static final int STATIC_ICMP        = 15;

  public static final int STATIC_OVERLOAD    = 95;
  public static final int STATIC_CHKPORTSCAN = 96;
  public static final int STATIC_CHECKSATAN  = 97;
  public static final int STATIC_LOGATTACK   = 98;
  public static final int STATIC_CHECKATTACK = 99;
  public static Vector staticLevels = new Vector(10,10);
  public static final int firstDynamic = 100; // level number of first dynamic level

/* This is a copy-paste template for creating new static levels:
    n = new Notification();
    n.levelName = "";
    n.levelText = "";
    staticLevels.addElement(n);
*/

  static { // default levels for templates
    Notification n;

    n = new Notification(STATIC_SOURCEROUTE);
    n.levelName = "SRCRTE";
    n.levelText = "report;\n"+
                  "message \"Suspicious IP packet has source route option set\";\n"+
                  "call "+STATIC_LOGATTACK+"; /* LOGATTACK: subroutine which checks for an attack */";
    staticLevels.addElement(n);

    n = new Notification(STATIC_SERVER);
    n.levelName = "SERVER";
    n.levelText = "message \"Incoming server connection\";\n"+
                  "call "+STATIC_OVERLOAD+"; /* OVERLOAD */";
    staticLevels.addElement(n);

    n = new Notification(STATIC_IDENT);
    n.levelName = "IDENT";
    n.levelText = "message \"Incoming ident request - connection closed\";";
    staticLevels.addElement(n);

    n = new Notification(STATIC_TRACEROUTE);
    n.levelName = "TRACERTE";
    n.levelText = "message \"Incoming traceroute\";\n"+
                  "report;\n"+
                  "reject with icmp_port_unreachable\n"+
                  "  udp from sourcehost to port 33434..33524\n"+
                  "  notification_level 0 timeout 60;";
    staticLevels.addElement(n);

    n = new Notification(STATIC_OUTGOING);
    n.levelName = "OUTGOING";
    n.levelText = "message \"Outgoing connection\";";
    staticLevels.addElement(n);

    n = new Notification(STATIC_PING);
    n.levelName = "PING";
    n.levelText = "message \"We have been pinged\";\n"+
                  "let satanping:sourcehost := satanping:sourcehost + 1 timeout 240;\n"+
                  "call "+STATIC_CHECKSATAN+"; /* CHKSATAN */ ";
    staticLevels.addElement(n);

    n = new Notification(STATIC_ICMP);
    n.levelName = "ICMP";
    n.levelText = "report;\n"+
                  "message \"Received suspicious ICMP packet\";\n"+
                  "call "+STATIC_LOGATTACK+"; /* LOGATTACK */";
    staticLevels.addElement(n);


    // Subroutines
    // -----------
    n = new Notification(STATIC_OVERLOAD);
    n.levelName = "OVERLOAD";
    n.levelText = "/* Check for attempt to overload the host */\n"+
                  "let connections:sourcehost := connections:sourcehost + 1 timeout 15;\n"+
                  "if connections:sourcehost > 50 then\n"+
                  "  spy;\n"+
                  "  report;\n"+
                  "  message \"Someone is trying to establish more than 50 connections\"\n"+
                  "          \"Blocking access for 30 seconds.\";\n"+
                  "  block all from sourcehost notification_level 0 timeout 30;\n"+
                  "  call "+STATIC_CHECKATTACK+"; /* CHKATTACK */\n"+
                  "endif;";
    staticLevels.addElement(n);

    n = new Notification(STATIC_CHKPORTSCAN);
    n.levelName = "CHKPRTSCN";
    n.levelText = "/* Check for port scanner */\n"+
                  "let oldattackport:sourcehost := attackport:sourcehost timeout 20;\n"+
                  "let attackport:sourcehost := destport timeout 600;\n"+
                  "if oldattackport:sourcehost != 0 then\n"+
                  "  if attackport:sourcehost != oldattackport:sourcehost then\n"+
                  "    let scan:sourcehost := scan:sourcehost + 1 timeout 600;\n"+
                  "    if scan:sourcehost > 2 then\n"+
                  "      message \"Someone is scanning ports\";\n"+
                  "      spy;\n"+
                  "      report;\n"+
                  "      let attack:sourcehost := attack:sourcehost + 12 timeout 600;\n"+
                  "      call "+STATIC_CHECKATTACK+"; /* CHKATTACK */\n"+
                  "    endif;\n"+
                  "  endif;\n"+
                  "endif;";
    staticLevels.addElement(n);

    n = new Notification(STATIC_CHECKSATAN);
    n.levelName = "CHKSATAN";
    n.levelText = "/* Check for SATAN */\n"+
                  "if satanping:sourcehost > 0 then\n"+
                  "  if satanrpc:sourcehost > 0 then\n"+
                  "    message \"SATAN attack - blocking access for 10 minutes\";\n"+
                  "    block all from sourcehost notification_level 0 timeout 600;\n"+
                  "    spy;\n"+
                  "    report;\n"+
                  "  endif;\n"+
                  "endif;";
    staticLevels.addElement(n);

    n = new Notification(STATIC_LOGATTACK);
    n.levelName = "LOGATTACK";
    n.levelText = "/* Log severe policy violation */\n"+
                  "let attack:sourcehost := attack:sourcehost + 1 timeout 600;\n"+
                  "call "+STATIC_CHECKATTACK+"; /* CHKATTACK */";
    staticLevels.addElement(n);

    n = new Notification(STATIC_CHECKATTACK);
    n.levelName = "CHKATTACK";
    n.levelText =  "/* Check if there is an attack and send e-mail */\n"+
                   "if attack:sourcehost > 12 then\n"+
                   "  report;\n"+
                   "  message \"Someone is trying to attack the network\";\n"+
                   "endif;\n"+
                   "if attack:sourcehost > 36 then\n"+
                   "  report;\n"+
                   "  message \"Someone is trying HARD to attack the network!\"\n"+
                   "          \"Blocking access for 10 minutes\";\n"+
                   "  block all from sourcehost notification_level 0 timeout 600;\n"+
                   "endif;";
    staticLevels.addElement(n);

  }

} // Notification

// Dialog box to edit a notification level
class NotificationDialog
	extends Dialog
	implements ActionListener {

  /**
   * @param co calling object, refresh() will be called after "OK"
   * @param p parent window
   * @param n notification level to be edited
   * @param md manage domain holding notification levels
   * @param nw if true, notification level will be removed from vector
   * in md when "Cancel" has been selected
   */   
  public NotificationDialog(Refreshable co, Frame p, Notification n, ManageDomain md, boolean nw) {
    super(p, "Notification Level", true);

    callingObject = co;
    parent = p;
    noti = n;
    mgDomain = md;
    newNoti = nw;

    GridBagLayout gbl = new GridBagLayout();
    GridBagConstraints gbc = new GridBagConstraints();
    gbc.fill = GridBagConstraints.BOTH;
    gbc.insets = new Insets(3,3,3,3);
    setLayout(gbl);

    lName = new TextField(n.levelName, 10);
    Utils.add_component(this, new Label("Name:"), gbl, gbc, 0, 0, 1, 1, 0, 0);
    Utils.add_component(this, lName, gbl, gbc, 1, 0, 1, 1, 0, 0);

    lText = new TextArea(n.levelText, 20, 80);
    lText.setFont(new Font("Courier", Font.PLAIN, 12));
    Utils.add_component(this, lText, gbl, gbc, 0, 1, 8, 20, 100, 100);

    gbc.fill = GridBagConstraints.NONE;
    gbc.anchor = GridBagConstraints.WEST;
    Panel pa = new Panel();
    OKbutton = new Button("OK");
		OKbutton.addActionListener(this);
    pa.add(OKbutton);
    Cancelbutton = new Button("Cancel");
		Cancelbutton.addActionListener(this);
    pa.add(Cancelbutton);
    Utils.add_component(this, pa, gbl, gbc, 0, 21, 8, 1, 0, 0);

    pack();
		addWindowListener(new NLAdapter());
  }

class NLAdapter extends WindowAdapter {

	public void WindowClosing(WindowEvent we) {
      if (isEnabled())
        dispose();
      return;
  }
} // class NLAdapter


  public void actionPerformed(ActionEvent ae) {
	Object source = ae.getSource();

		// Process OKbutton
    if (source == OKbutton) {
      // don't save changes in static notification levels
      if (noti.levelNumber < Notification.firstDynamic) {
        dispose();
        return;
      }

      String str = lName.getText().trim().toUpperCase();
      Enumeration en = mgDomain.NLevels.elements();
      Notification n;
      while (en.hasMoreElements()) {
        n = (Notification)en.nextElement();
        if (n != noti && n.levelName.equals(str)) {
          UserDialog.ErrorBox("Level "+str+" already exists!");
          return;
        }
      }
      noti.levelName = str;
      noti.levelText = lText.getText();
      if (callingObject != null)
        callingObject.refresh();
      dispose();
      return;
    }

		//Process Cancelbutton
    else if (source == Cancelbutton) {
      if (newNoti)
        mgDomain.NLevels.removeElement(noti);
      dispose();
      return;
    }
  } // actionPerformed

  private Refreshable callingObject;
  private Frame parent;
  private Notification noti;
  private ManageDomain mgDomain;
  private boolean newNoti;

  private TextField lName;
  private TextArea lText;
  private Button OKbutton;
  private Button Cancelbutton;
} // NotificationDialog
